import torch
import numpy as np
import xarray as xr
import colorama as ca
import collections.abc

def check_cuda():
    p_hint(f'CUDA availability: {torch.cuda.is_available()}')
    p_hint(f'Available GPUs: {torch.cuda.device_count()}')

def p_header(text):
    print(ca.Fore.CYAN + ca.Style.BRIGHT + text + ca.Style.RESET_ALL)

def p_hint(text):
    print(ca.Fore.LIGHTBLACK_EX + ca.Style.BRIGHT + text + ca.Style.RESET_ALL)

def p_success(text):
    print(ca.Fore.GREEN + ca.Style.BRIGHT + text + ca.Style.RESET_ALL)

def p_fail(text):
    print(ca.Fore.RED + ca.Style.BRIGHT + text + ca.Style.RESET_ALL)

def p_warning(text):
    print(ca.Fore.YELLOW + ca.Style.BRIGHT + text + ca.Style.RESET_ALL)

def gcd(lat1, lon1, lat2, lon2, radius=6378.137):
    ''' 2D Great Circle Distance [km]

    Args:
        radius (float): Earth radius
    '''
    # Convert degrees to radians
    lon1, lat1, lon2, lat2 = map(np.radians, [lon1, lat1, lon2, lat2])
    dlat, dlon = lat2 - lat1, lon2 - lon1
    a = np.sin(dlat / 2)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2)**2
    c = 2 * np.arcsin(np.sqrt(a))
    dist = radius * c
    return dist

def states2ds(states, ds):
    original_shapes = {vn: ds[vn].shape for vn in ds.data_vars}
    original_dims = {vn: ds[vn].dims for vn in ds.data_vars}
    original_coords = {vn: ds[vn].coords for vn in ds.data_vars}
    # print(original_shapes)
    # print(original_dims)
    # print(original_coords)
    ds_out = xr.Dataset()
    start_loc = 0
    for vn in ds.data_vars:
        if 'ens' in original_dims[vn]:
            end_loc = start_loc + int(np.prod(original_shapes[vn][:-1]))
            # p_hint(f'{np.prod(original_shapes[vn][:-1]) = }')
        else:
            end_loc = start_loc + int(np.prod(original_shapes[vn]))
            # p_hint(f'{np.prod(original_shapes[vn]) = }')
        # print(vn, start_loc, end_loc)

        # p_hint(f'{vn = }')
        # p_hint(f'{start_loc = }')
        # p_hint(f'{end_loc = }')
        # p_hint(f'{np.shape(states) = }')
        # p_hint(f'{np.shape(states[start_loc:end_loc]) = }')
        data = states[start_loc:end_loc].reshape(original_shapes[vn])
        nan_mask = np.isnan(ds[vn].values)
        data[nan_mask] = np.nan

        ds_out[vn] = xr.DataArray(
            data,
            dims=original_dims[vn],
            coords=original_coords[vn],
        )
        start_loc = end_loc
        ds_out[vn].attrs = ds[vn].attrs

    return ds_out
        

# def gcd_3d(loc1, loc2, radius=6371.0):
#     ''' 3D Great Circle Distance [km]

#     Args:
#         loc1 (tuple): lat1 [degree], lon1 [degree], depth1 [km]
#         loc2 (tuple): lat2 [degree], lon2 [degree], depth2 [km]
#         radius (float): Earth radius
#     '''
#     lat1, lon1, depth1 = loc1
#     lat2, lon2, depth2 = loc2

#     # Convert degrees to radians
#     lat1, lon1 = np.radians(lat1), np.radians(lon1)
#     lat2, lon2 = np.radians(lat2), np.radians(lon2)
    
#     # Calculate radial distances (Earth's radius minus depth)
#     r1 = radius - depth1
#     r2 = radius - depth2
    
#     # Compute central angle component
#     central_angle = np.sin(lat1) * np.sin(lat2) + np.cos(lat1) * np.cos(lat2) * np.cos(lon2 - lon1)
    
#     # Compute the 3D distance
#     distance_3d = np.sqrt(r1**2 + r2**2 - 2 * r1 * r2 * central_angle)
    
#     return distance_3d

def str2list(s, sep=','):
    l = [int(ss.strip()) for ss in s.split(sep)]
    return l

def lon360(lon180):
    return np.mod(lon180, 360)

def lon180(lon360):
    return np.mod(lon360 + 180, 360) - 180

def update_dict(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_dict(d.get(k, {}), v)
        else:
            d[k] = v
    return d


import numpy as np
import xarray as xr
import xesmf as xe
from tqdm import tqdm
from scipy.stats import norm
from tqdm import tqdm
from copy import deepcopy

from . import psm
from . import utils

class PriorMember:
    def __init__(self, ds):
        if isinstance(ds, xr.DataArray): ds = ds.to_dataset()
        self.ds = ds
        self.vns = list(ds.data_vars)

    def gen_samples_Gaussian(self, local_sigma:dict, global_sigma:dict, nens:int=100, seed:int=0):
        ''' Generate samples following Gaussian

        Args:
            sigma (dict): Dictionary with standard deviation (sigma) for each variable.
            nens (int): Number of ensemble members to generate.
            seed (int): Seed for reproducibility.
        '''
        rng = np.random.default_rng(seed)
        self.samples = xr.Dataset()
        for vn in self.vns:
            mean = self.ds[vn].values
            samples_shape = (*mean.shape, nens)
            global_perturbation = norm.rvs(loc=0, scale=global_sigma[vn], size=nens, random_state=rng)
            local_perturbation = norm.rvs(loc=0, scale=local_sigma[vn], size=samples_shape, random_state=rng)
            samples = mean[..., np.newaxis] + global_perturbation + local_perturbation
            samples_da = xr.DataArray(samples, dims=(*self.ds[vn].dims, 'ens'), coords=self.ds[vn].coords)
            samples_da.attrs = self.ds[vn].attrs
            self.samples[vn] = samples_da

    def gen_samples_bootstrap(self, nens:int=30, clim_yrs:int=50, seed:int=0, dim='time'):
        ''' Generate samples from the prior pool

        Args:
        '''
        nt = len(self.ds[dim])
        pool_idx = list(range(nt))
        sample_list = []
        for i in range(nens):
            seed += 1
            rng = np.random.default_rng(seed)
            sample_idx = rng.choice(pool_idx, size=clim_yrs, replace=False)
            sample = self.ds.isel({dim: sample_idx}).mean(dim)
            sample_list.append(sample)

        samples = xr.concat(sample_list, dim=dim)
        self.samples = xr.Dataset(samples).rename({dim: 'ens'}).transpose(..., 'ens')


class Prior:
    def __init__(self, members):
        if not isinstance(members, list): members = [members]
        ds_list = []
        for m in members:
            if hasattr(m, 'samples'):
                ds_list.append(m.samples)
            else:
                if 'ens' not in m.ds.dims:
                    ds_list.append(m.ds.expand_dims({'ens': 1}))
                else:
                    ds_list.append(m.ds)

        self.ds = xr.concat(ds_list, dim='ens').transpose(..., 'ens')
        self.nens = len(self.ds.ens)
        self.nvar = len(self.ds.data_vars)

    def regrid(self, ds_template=None, dlat=1, dlon=1, verbose=False, **kws):
        if ds_template is None:
            self.ds_rgd = xr.Dataset()
            for vn in tqdm(self.ds.data_vars, desc=f'Regridding variables to {dlat}x{dlon}'):
                self.ds_rgd[vn] = self.ds.x[vn].x.regrid(dlat=dlat, dlon=dlon)
                if verbose: utils.p_success(f'>>> Prior.ds_rgd["{vn}"] created')
        else:
            _kws = {
                'periodic': True,
                'reuse_weights': True,
            }
            _kws.update(kws)
            regridder = xe.Regridder(self.ds, ds_template, method='bilinear', **kws)
            self.ds_rgd = regridder(self.ds)
            for vn in self.ds_rgd.data_vars:
                self.ds_rgd[vn].attrs = ds_template[vn].attrs.copy()
                if verbose: utils.p_success(f'>>> Prior.ds_rgd["{vn}"] created')

        self.ds_rgd = self.ds_rgd.transpose(..., 'ens')


    def annualize(self, months=list(range(1, 13))):
        return self.ds.sel(month=months).mean('month')

    def inflate(self, factor=2):
        self.ds_raw = self.ds.copy()
        ens_mean = self.ds.mean('ens')
        ens_pert = self.ds - self.ds.mean('ens')
        inflated_pert = ens_pert * factor
        self.ds = ens_mean + inflated_pert

    def copy(self):
        return deepcopy(self)


    @property
    def X(self):
        res = []
        for vn in self.ds.data_vars:
            res.append(self.ds[vn].values.reshape(-1, self.nens))

        res = np.concatenate(res, axis=0)
        return res

    # def get_Y(self, obs, **fwd_kws):
    #     self.obs_assim = obs.copy()
    #     lats = obs.df['lat'].values
    #     lons = obs.df['lon'].values
    #     pids = obs.df['pid'].values
    #     depths = obs.df['depth'].values
    #     if 'clean' in obs.df.columns: cleans = obs.df['clean'].values
    #     if 'species' in obs.df.columns: specs = obs.df['species'].values

    #     psms = obs.df['psm'].values
    #     psm_names = list(set(psms))

    #     pseudo_obs = np.empty((len(obs.df), self.nens))

    #     # Loop over PSM types (psm_names)
    #     for psm_name in psm_names:
    #         mask = psms == psm_name
    #         idx = np.where(mask)[0]
    #         if np.any(mask):
    #             lat_lon_pairs = xr.Dataset({
    #                 'lat': (('obs',), lats[mask]),
    #                 'lon': (('obs',), lons[mask]),
    #             })
    #             self.clim_proxy_locs = xr.Dataset()
    #             for vn in self.ds_rgd.data_vars:
    #                 filled_da = self.ds_rgd[vn].ffill(dim='lat').bfill(dim='lat').ffill(dim='lon').bfill(dim='lon')
    #                 self.clim_proxy_locs[vn] = filled_da.sel(
    #                     lat=lat_lon_pairs['lat'],
    #                     lon=lat_lon_pairs['lon'],
    #                     method='nearest',
    #                 ).transpose(..., 'ens')

    #             for i in tqdm(range(len(idx)), desc=f'>>> Looping over sites w/ PSM - {psm_name}'):
    #                 pid = pids[idx[i]]
    #                 lat = lats[idx[i]]
    #                 lon = lons[idx[i]]
    #                 depth = depths[idx[i]]
    #                 if np.isnan(depth): depth = 0

    #                 obs_meta = {
    #                     'pid': pid,
    #                     'lat': lat,
    #                     'lon': lon,
    #                     'depth': depth,
    #                 }
    #                 if 'clean' in obs.df.columns:
    #                     clean = cleans[idx[i]]
    #                     if np.isnan(clean): clean = 0
    #                     obs_meta['clean'] = clean

    #                 if 'species' in obs.df.columns:
    #                     species = specs[idx[i]]
    #                     if not isinstance(species, str): species = 'all'
    #                     obs_meta['species'] = species

    #                 mdl = psm.__dict__[psm_name](obs_meta, self.clim_proxy_locs.isel({'obs': i}))
    #                 _fwd_kws = {}
    #                 _fwd_kws[psm_name] = {}
    #                 if psm_name in fwd_kws:
    #                     _fwd_kws[psm_name].update(fwd_kws[psm_name])

    #                 res = mdl.forward(**_fwd_kws[psm_name])
    #                 if res is None:
    #                     utils.p_warning(f'>>> Dropping proxy: {pid}')
    #                     self.obs_assim.df = obs.df.drop(obs.df[obs.df['pid'] == pid].index)
    #                     pseudo_obs[idx[i]] = np.nan
    #                 else:
    #                     pseudo_obs[idx[i]] = res

    #     self.obs_assim.nobs = len(self.obs_assim.df)
    #     pseudo_obs = pseudo_obs[~np.isnan(pseudo_obs).any(axis=1)]
    #     self.Y = pseudo_obs
    #     self.obs_assim.df['Ym'] = self.Y.mean(axis=1)

    # def get_Y(self, obs, depth_name='z_t', nearest_kws={}, **fwd_kws):
    #     self.obs_assim = obs.copy()
    #     psm_names = set(obs.df['psm_name'])
    #     clim_vns = list({
    #         vn for psm_name in psm_names
    #         for vn in psm.__dict__[psm_name]().clim_vns
    #         if vn in self.ds.data_vars
    #     })

    #     self.ds_proxy_locs = xr.Dataset()
    #     _nearest_kws = {}
    #     _nearest_kws.update(nearest_kws)
    #     for vn in clim_vns:
    #         da = self.ds[vn].isel({depth_name: 0}) if depth_name in self.ds[vn].dims else self.ds[vn]
    #         self.ds_proxy_locs[vn] = da.x.nearest2d(
    #             lat=obs.df['lat'],
    #             lon=obs.df['lon'],
    #             **_nearest_kws
    #         ).transpose(..., 'ens')

    #     pseudo_obs = np.empty((len(obs.df), self.nens))
    #     for i, (pid, rec) in tqdm(enumerate(obs.records.items()), total=obs.nobs, desc='Looping over records'):
    #         self.obs_assim.records[pid].psm = mdl = psm.__dict__[rec.data.psm_name](rec)
    #         if 'month' in self.ds_proxy_locs.dims:
    #             mdl.clim = self.ds_proxy_locs.isel({'site': i}).sel(month=rec.data.seasonality).mean(dim='month')
    #         else:
    #             mdl.clim = self.ds_proxy_locs.isel({'site': i})

    #         for vn in clim_vns:
    #             if vn in mdl.clim and mdl.clim[vn].isnull().any():
    #                 raise ValueError(f'NaN values detected in input climate for forward modeling of: {pid}')

    #         _fwd_kws = {}
    #         _fwd_kws[rec.data.psm_name] = {}
    #         if rec.data.psm_name in fwd_kws:
    #             _fwd_kws[rec.data.psm_name].update(fwd_kws[rec.data.psm_name])
    #         mdl.forward(**_fwd_kws[rec.data.psm_name])
    #         if mdl.output is None:
    #             utils.p_warning(f'>>> Dropping proxy: {pid}')
    #             self.obs_assim.df = obs.df.drop(obs.df[obs.df['pid'] == pid].index)
    #             pseudo_obs[i] = np.nan
    #         else:
    #             pseudo_obs[i] = mdl.output

    #     self.obs_assim.nobs = len(self.obs_assim.df)
    #     pseudo_obs = pseudo_obs[~np.isnan(pseudo_obs).any(axis=1)]
    #     self.Y = pseudo_obs
    #     self.obs_assim.df['Ym'] = self.Y.mean(axis=1)
    def get_Y(self, obs):
        self.Y = np.empty((len(obs.df), self.nens))
        for idx, row in obs.df.iterrows():
            self.Y[idx, :] = row['pseudo']

    # def get_Y(self, obs, nearest_valid_radius=5, **fwd_kws):
    #     self.obs_assim = obs.copy()
    #     pseudo_obs = np.empty((len(obs.df), self.nens))

    #     psm_names = set(obs.df['psm_name'])
    #     clim_vns = list({
    #         vn for psm_name in psm_names
    #         for vn in psm.__dict__[psm_name]().clim_vns
    #         if vn in self.ds_rgd.data_vars
    #     })

    #     lat_lon_pairs = xr.Dataset({
    #         'lat': (('sites',), obs.df['lat'].values),
    #         'lon': (('sites',), obs.df['lon'].values),
    #     })
    #     self.ds_proxy_locs = xr.Dataset()
    #     for vn in clim_vns:
    #         # filled_da = self.ds_rgd[vn].ffill(dim='lon').bfill(dim='lon').ffill(dim='lat').bfill(dim='lat')
    #         # ds_proxy_locs[vn] = filled_da.sel(
    #         #     lat=lat_lon_pairs['lat'],
    #         #     lon=lat_lon_pairs['lon'],
    #         #     method='nearest',
    #         # ).transpose(..., 'ens')

    #         self.ds_proxy_locs[vn] = self.ds_rgd[vn].x.nearest2d(
    #             lat=lat_lon_pairs['lat'],
    #             lon=lat_lon_pairs['lon'],
    #             r=nearest_valid_radius,
    #             extra_dim='ens',
    #         ).transpose(..., 'ens')

    #     if 'sites' not in self.ds_proxy_locs.dims:
    #         self.ds_proxy_locs = self.ds_proxy_locs.expand_dims({'sites': [0]})

    #         # if ds_proxy_locs[vn].isnull().any():
    #         #     for idx in obs.df.index:
    #         #         if ds_proxy_locs[vn].sel(sites=idx).isnull().any():
    #         #             utils.p_warning(f"NaN detected for {vn}: {obs.df.iloc[idx][['pid', 'lat', 'lon']].values}")
    #         #             print(ds_proxy_locs[vn].sel(sites=idx).dims)
    #         #             print(ds_proxy_locs[vn].sel(sites=idx).values)
    #         #             utils.p_warning('------------------------------------')
    #         #     raise ValueError('Some of the nearest gridcell values are NaN.')

    #     # nearest_lats, nearest_lons = [], []
    #     for i, (pid, rec) in tqdm(enumerate(obs.records.items()), total=obs.nobs, desc='Looping over records'):
    #         # nearest_clim = self.ds_proxy_locs.isel({'sites': i}).sel(month=rec.data.seasonality).mean(dim='month')
    #         # nearest_lat = nearest_clim.lat.values.mean()
    #         # nearest_lon = nearest_clim.lon.values.mean()
    #         # nearest_lats.append(nearest_lat)
    #         # nearest_lons.append(nearest_lon)
    #         # rec.data.lat = nearest_lat
    #         # rec.data.lon = nearest_lon

    #         mdl = psm.__dict__[rec.data.psm_name](rec)
    #         if 'month' in self.ds_proxy_locs.dims:
    #             mdl.record.clim = self.ds_proxy_locs.isel({'sites': i}).sel(month=rec.data.seasonality).mean(dim='month')
    #         else:
    #             mdl.record.clim = self.ds_proxy_locs.isel({'sites': i})

    #         for vn in clim_vns:
    #             if mdl.record.clim[vn].isnull().any():
    #                 # print(i, ds_proxy_locs[vn].isel({'sites': i}))
    #                 # print(ds_proxy_locs[vn].isel({'sites': i, 'ens': 6}))
    #                 # print(ds_proxy_locs.sel(month=rec.data.seasonality).mean(dim='month')[vn].values[i])
    #                 # print(vn, rec.data.pid, rec.data.lat, rec.data.lon, rec.data.seasonality)
    #                 # print(mdl.record.clim[vn].values)
    #                 raise ValueError(f'NaN values detected in input climate for forward modeling of: {pid}')

    #         _fwd_kws = {}
    #         _fwd_kws[rec.data.psm_name] = {}
    #         if rec.data.psm_name in fwd_kws:
    #             _fwd_kws[rec.data.psm_name].update(fwd_kws[rec.data.psm_name])
    #         mdl.forward(**_fwd_kws[rec.data.psm_name])
    #         if mdl.output is None:
    #             utils.p_warning(f'>>> Dropping proxy: {pid}')
    #             self.obs_assim.df = obs.df.drop(obs.df[obs.df['pid'] == pid].index)
    #             pseudo_obs[i] = np.nan
    #         else:
    #             pseudo_obs[i] = mdl.output

    #         self.obs_assim.records[pid].psm = mdl  # for debugging purposes 

    #     self.obs_assim.nobs = len(self.obs_assim.df)
    #     # self.obs_assim.df['lat'] = nearest_lats
    #     # self.obs_assim.df['lon'] = nearest_lons
    #     pseudo_obs = pseudo_obs[~np.isnan(pseudo_obs).any(axis=1)]
    #     self.Y = pseudo_obs
    #     self.obs_assim.df['Ym'] = self.Y.mean(axis=1)

    # def get_dist(self, obs, s=1):
    #     # Extract grid latitudes and longitudes as 2D arrays
    #     lat_grid = self.ds[self.lat_name].values  # shape: (nlat, nlon)
    #     lon_grid = self.ds[self.lon_name].values  # shape: (nlat, nlon)

    #     if lat_grid.ndim == 1 and lon_grid.ndim == 1:
    #         # If lat and lon are 1D, create a meshgrid
    #         lon_grid, lat_grid = np.meshgrid(lon_grid, lat_grid)

    #     # Flatten the grid arrays to 1D
    #     lat_grid_flat = lat_grid.ravel()  # shape: (nlat * nlon,)
    #     lon_grid_flat = lon_grid.ravel()  # shape: (nlat * nlon,)

    #     # Get the observation lat/lon as a 2D array
    #     lats2 = obs.df['lat'].values  # shape: (nobs,)
    #     lons2 = obs.df['lon'].values  # shape: (nobs,)

    #     # Broadcast the grid cells to all observation points
    #     lats1 = np.repeat(lat_grid_flat, obs.nobs)  # shape: (nlat * nlon * nobs,)
    #     lons1 = np.repeat(lon_grid_flat, obs.nobs)  # shape: (nlat * nlon * nobs,)

    #     # Repeat observation points for every grid point
    #     lats2 = np.tile(lats2, len(lat_grid_flat))  # shape: (nlat * nlon * nobs,)
    #     lons2 = np.tile(lons2, len(lon_grid_flat))  # shape: (nlat * nlon * nobs,)

    #     dist0 = utils.gcd(lats1, lons1, lats2, lons2).reshape((-1, obs.nobs))

    #     if hasattr(self, 'nz'):
    #         # 3D localization
    #         s = (np.ones(self.nz)*s).reshape(-1, 1, 1)
    #         dist1 = (dist0[None, :, :] * s).reshape((-1, obs.nobs))
    #         self.dist = dist1[np.newaxis, :].repeat(self.nvar, axis=0).reshape(-1, obs.nobs)
    #     else:
    #         self.dist = dist0[None, :, :].repeat(self.nvar, axis=0).reshape(-1, obs.nobs)
    def set_hcoords(self, hcoords:dict):
        self.hcoords = hcoords

    def get_dist(self, obs, vertical_dims=['z_t', 'lev'], s=1):
        all_dists = []
        for vn in tqdm(self.ds.data_vars, desc='Processing variables'):
            lat_name, lon_name = self.hcoords[vn] if hasattr(self, 'hcoords') else ('lat', 'lon')
            lats, lons = self.ds[vn].coords[lat_name].values, self.ds[vn].coords[lon_name].values
            if lat_name not in self.ds[vn].dims:
                if len(lats.shape) == 2:
                    # POP-like curvilinear grid
                    lats = lats.ravel()
                    lons = lons.ravel()
            else:
                # regular lat-lon grid
                lons, lats = np.meshgrid(lons, lats)
                lats = lats.ravel()
                lons = lons.ravel()

            lats1 = np.repeat(lats, obs.nobs)
            lons1 = np.repeat(lons, obs.nobs)
            lats2 = np.tile(obs.df['lat'].values, len(lats))
            lons2 = np.tile(obs.df['lon'].values, len(lons))

            dist = utils.gcd(lats1, lons1, lats2, lons2).reshape((-1, obs.nobs))

            vdims = list(set(self.ds[vn].dims) & set(vertical_dims))
            if len(vdims) == 1:
                nz = len(self.ds[vn][vdims[0]])
                s_arr = np.reshape(np.ones(nz) * s, (-1, 1, 1))
                dist = (dist[None, :, :] * s_arr).reshape((-1, obs.nobs))

            all_dists.append(dist)
        
        self.dist = np.concatenate(all_dists, axis=0)




# get the version
from importlib.metadata import version

__version__ = version('mlda')
import warnings
warnings.filterwarnings("ignore", category=UserWarning)

from . import utils, nn, enkf, cesm
from .prior import Prior, PriorMember
from .obs import Obs




from copy import deepcopy
import numpy as np
import pandas as pd
import xarray as xr
import x4c
import plotly.express as px
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator
from matplotlib.colors import BoundaryNorm
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from tqdm import tqdm

from . import psm
from . import utils
class Obs:
    def __init__(self, df:pd.DataFrame):
        self.df = df.reset_index(drop=True)   # with columns: pid, lat, lon, time, value, psm_name, R, type
        self.df['lon'] = pd.to_numeric(self.df['lon'], errors='coerce')
        self.df['lon'] = (self.df['lon'] + 360) % 360

    def setup(self):
        self.df = self.df.reset_index(drop=True)
        self.nobs = len(self.df)
        self.pids = self.df['pid'].values
        self.to_ds()
        self.records = {}
        for pid in self.pids:
            self.records[pid] = self[pid]

    @property
    def y(self):
        # return self.df['value'].values[..., np.newaxis]
        return self.ds.to_dataframe().values

    @property
    def y_locs(self):
        return self.df[['lat', 'lon']].values

    @property
    def R(self):
        return np.diag(self.df['R'].values)

    def copy(self):
        return deepcopy(self)

    def __getitem__(self, pid:str):
        mask = self.df['pid'] == pid
        row = self.df[mask].iloc[0]
        rec = ProxyRecord(row)
        return rec
    
    def get_dist(self):
        lats = self.df['lat'].values
        lons = self.df['lon'].values
        lat1, lat2 = np.meshgrid(lats, lats)
        lon1, lon2 = np.meshgrid(lons, lons)
        self.dist = utils.gcd(lat1, lon1, lat2, lon2)
        return self.dist

    def get_clim(self, prior, depth_name='z_t', ens_name='ens', mode='each_member', **nearest_kws):
        psm_names = set(self.df['psm_name'])
        clim_vns = list({
            vn for psm_name in psm_names
            for vn in psm.__dict__[psm_name]().clim_vns
            if vn in prior.ds.data_vars
        })

        self.clim = xr.Dataset()
        if mode == 'each_member':
            ens_vals = prior.ds[ens_name].values
            for vn in clim_vns:
                da = prior.ds[vn]
                if depth_name in da.dims: da = da.isel({depth_name: 0})
                member_list = []
                for i, ens_val in enumerate(ens_vals):
                    da_ens = da.sel({ens_name: ens_val})
                    result = da_ens.x.nearest2d(
                        lat=self.df['lat'],
                        lon=self.df['lon'],
                        **nearest_kws,
                    )
                    result = result.expand_dims({ens_name: [ens_val]})  # ensure ens dimension exists
                    member_list.append(result)

                result_all = xr.concat(member_list, dim=ens_name)
                self.clim[vn] = result_all.transpose(..., ens_name)
        else:
            for vn in clim_vns:
                da = prior.ds[vn]
                if depth_name in da.dims: da = da.isel({depth_name: 0})
                self.clim[vn] = da.x.nearest2d(
                    lat=self.df['lat'],
                    lon=self.df['lon'],
                    **nearest_kws,
            ).transpose(..., ens_name)

        self.clim.coords['site'] = self.df['pid']

    def get_pseudo(self, psm_colname='psm_name', **fwd_kws):
        self.df['pseudo'] = object
        for idx, row in self.df.iterrows():
            pid = row['pid']
            psm_name = row[psm_colname]
            if psm_name in fwd_kws:
                _fwd_kws = fwd_kws[psm_name]
            else:
                _fwd_kws = {}

            self.records[pid].psm = mdl = psm.__dict__[psm_name](self.records[pid])
            mdl.clim = self.clim.sel(site=pid)
            if 'month' in mdl.clim.dims:
                mdl.clim = mdl.clim.sel(month=self.records[pid].data.seasonality).mean(dim='month')

            self.df.at[idx, 'pseudo'] = mdl.forward(**_fwd_kws)
            self.df.at[idx, 'Ym'] =self.df.at[idx, 'pseudo'].mean()

    def to_ds(self):
        ds = xr.Dataset()
        for idx, row in self.df.iterrows():
            da = xr.DataArray(
                data=row['value'],
                coords={
                    'time': row['time'],
                },
                name=row['pid'],
            )
            for col in row.index:
                if col not in ['pid', 'time', 'value']:
                    da.attrs[col] = row[col]

            ds[row['pid']] = da
        self.ds = ds
        if 'time' not in self.ds.dims: self.ds['time'] = self.ds.time.expand_dims(dim='time')
        return self.ds

    def plot(self, t_idx=None, return_im=False, da_target=None, **kws):
        # useful for pseudoproxy experiments
        if t_idx is None:
            if len(da_target.dims) > 1 and 'time' in da_target.dims:
                da_target = da_target.mean('time')
        else:
            da_target = da_target.isel(time=t_idx)

        fig_ax_im = da_target.x.plot(return_im=True, **kws)
        if len(fig_ax_im) == 3:
            fig, ax, im = fig_ax_im
        elif len(fig_ax_im) == 2:
            ax, im
        else:
            raise ValueError('Unexpected number of return values from x4c plot.')

        lats = self.df['lat'].values
        lons = self.df['lon'].values
        vals = np.array([v for v in self.df['value'].values])
        if t_idx is None:
            if len(vals.shape) > 1 and len(vals) > 1:
                vals = vals.mean(axis=-1)
        else:
            vals = vals[:, t_idx]

        _scatter_kws = {
            's': 100,
            'cmap': im.cmap,
            'norm': im.norm,
            'transform': ccrs.PlateCarree(),
            'edgecolor': 'k',
            'zorder': 99,
        }
        ax.scatter(lons, lats, c=vals, **_scatter_kws)

        if 'fig' in locals():
            return (fig, ax, im) if return_im else (fig, ax)
        else:
            return (ax, im) if return_im else ax

    def plotly(self, **kws):
        ''' Plot the database on an interactive map utilizing Plotly
        '''
        _kws = {
            'lat': 'lat',
            'lon': 'lon',
            'color': 'type',
            'hover_name': 'pid',
            'hover_data': self.df.columns.drop(['lat', 'lon', 'type', 'pid', 'time', 'value'], errors='ignore'),
            'projection': 'natural earth',
        }
        _kws.update(kws)
        return px.scatter_geo(self.df, **_kws)


class ProxyRecord:
    def __init__(self, data:pd.Series):
        self.data = data.copy()
        if 'time' in data: self.data['time'] = np.array(data['time'])
        if 'value' in data: self.data['value'] = np.array(data['value']) 

        if 'seasonality' in data:
            if isinstance(data['seasonality'], str):
                self.data['seasonality'] = utils.str2list(data['seasonality'])
            elif isinstance(data['seasonality'], list):
                self.data['seasonality'] = data['seasonality']
            else:
                raise ValueError('Wrong seasonality type; should be a string or a list.')


#     def get_clim(self, clim_ds, vns:list=None, verbose=False):
#         if vns is None:
#             vns = clim_ds.data_vars
#         else:
#             vns = [vn for vn in vns if vn in clim_ds.data_vars]

#         self.clim = xr.Dataset()
#         for vn in vns:
#             self.clim[vn] = clim_ds[vn].x.nearest2d(
#             # filled_da = clim_ds[vn].ffill(dim='lon').bfill(dim='lon').ffill(dim='lat').bfill(dim='lat')
#             # self.clim[vn] = filled_da.sel(
#                 lat=self.data.lat,
#                 lon=self.data.lon,
#                 method='nearest',
#             ).sel(month=self.data.seasonality).mean(dim='month')
#             if verbose: utils.p_success(f'>>> ProxyRecord.clim["{vn}"] created')

#         self.clim.attrs['seasonality'] = self.data.seasonality




import os
import xarray as xr
import numpy as np
import torch
import matplotlib.pyplot as plt
from matplotlib.colors import BoundaryNorm

from sklearn.metrics import (
    r2_score,
    mean_squared_error,
    mean_absolute_error,
    mean_absolute_percentage_error,
)

from .. import utils

class Predictor:
    def __init__(self, model, data, ckpt_fpath):
        self.model = model
        self.data = data
        self.ckpt_fpath = ckpt_fpath
        self.ckpt = torch.load(self.ckpt_fpath, weights_only=True)
        self.model.load_state_dict(self.ckpt['state_dict'])

    def run(self, pred_fpath=None, truth_fpath=None, set='test'):
        if set == 'train':
            dataset = self.data.train_valid_set
            idx_start = self.data.train_valid_idx[0]
            idx_end = self.data.train_valid_idx[-1]
        elif set == 'test':
            dataset = self.data.test_set
            idx_start = self.data.test_idx[0]
            idx_end = self.data.test_idx[-1]

        self.y = [data['target'].detach().numpy() for data in dataset]
        self.y_hat = [self.model.forward(data['obs_vals'].unsqueeze(0))[0].squeeze(0).detach().numpy() for data in dataset]

        self.da_truth = xr.DataArray()
        self.da_pred = xr.Dataset()

        ds_template = self.data.target_da[idx_start:idx_end+1]
        self.da_truth = ds_template.copy()
        self.da_truth.values = np.array(self.y)
        self.da_pred = ds_template.copy()
        self.da_pred.values = np.array(self.y_hat)

        if truth_fpath is not None:
            if os.path.exists(truth_fpath): os.remove(truth_fpath)
            dirpath = os.path.dirname(truth_fpath)
            if not os.path.exists(dirpath): os.makedirs(dirpath, exist_ok=True)
            self.ds_truth.to_netcdf(truth_fpath)
            utils.p_success(f'Truth saved at: "{truth_fpath}"')

        if pred_fpath is not None:
            if os.path.exists(pred_fpath): os.remove(pred_fpath)
            dirpath = os.path.dirname(pred_fpath)
            if not os.path.exists(dirpath): os.makedirs(dirpath, exist_ok=True)
            self.ds_pred.to_netcdf(pred_fpath)
            utils.p_success(f'Prediction saved at: "{pred_fpath}"')

    def plot(self, t_idx=None, figsize=(16, 10), nrow=3, ncol=2, ax_loc=None, projs=None, projs_kws=None, wspace=0.1, hspace=0.2, **kws):
        try:
            import x4c
        except ImportError:
            raise ImportError('x4c is required for this method. Please install x4c via: pip install x4c-exp.')

        x4c.set_style('journal_spines', font_scale=1.2)
        ax_loc={
            'Truth': (0, 0),
            'Pred': (1, 0),
            'Diff': (2, 0),
            'corr': (0, 1),
            'R2': (1, 1),
            'RMSE': (2, 1),
        } if ax_loc is None else ax_loc

        projs={
            'Truth': 'Robinson',
            'Pred': 'Robinson',
            'Diff': 'Robinson',
            'corr': 'Robinson',
            'R2': 'Robinson',
            'RMSE': 'Robinson',
        } if projs is None else projs

        projs_kws={
            'Truth': {'central_longitude': 180},
            'Pred': {'central_longitude': 180},
            'Diff': {'central_longitude': 180},
            'corr': {'central_longitude': 180},
            'R2': {'central_longitude': 180},
            'RMSE': {'central_longitude': 180},
        } if projs_kws is None else projs_kws

        if t_idx is None:
            if len(self.da_truth.dims) > 1 and 'time' in self.da_truth.dims:
                da_truth = self.da_truth.mean('time')
        else:
            ax_loc.pop('corr')
            ax_loc.pop('R2')
            ax_loc.pop('RMSE')
            projs.pop('corr')
            projs.pop('R2')
            projs.pop('RMSE')
            projs_kws.pop('corr')
            projs_kws.pop('R2')
            projs_kws.pop('RMSE')
            figsize = (16, 10)
            ncol = 1
            da_truth = self.da_truth.isel(time=t_idx)

        if t_idx is None:
            if len(self.da_pred.dims) > 1 and 'time' in self.da_pred.dims:
                da_pred = self.da_pred.mean('time')
        else:
            da_pred = self.da_pred.isel(time=t_idx)

        fig, ax = x4c.visual.subplots(
            nrow=nrow, ncol=ncol,
            ax_loc=ax_loc,
            projs=projs,
            projs_kws=projs_kws,
            figsize=figsize,
            wspace=wspace,
            hspace=hspace,
        )

        _kws_truth = kws.copy()
        _kws_truth.update({
            'title': 'Truth',
        })
        da_truth.x.plot(ax=ax['Truth'], **_kws_truth)

        _kws_pred = kws.copy()
        _kws_pred.update({
            'title': 'Prediction',
        })
        da_pred.x.plot(ax=ax['Pred'], **_kws_pred)

        _kws_diff = kws.copy()
        _kws_diff.update({
            'levels': np.linspace(-1, 1, 11),
            'cbar_kwargs': {
                'ticks': np.linspace(-1, 1, 11),
            },
            'title': 'Prediction $-$ Truth',
        })
        diff = da_pred - da_truth
        diff.x.plot(ax=ax['Diff'], **_kws_diff)

        if t_idx is None:
            _kws_corr = kws.copy()
            _kws_corr.update({
                'levels': np.linspace(-1, 1, 11),
                'cbar_kwargs': {
                    'ticks': np.linspace(-1, 1, 11),
                    'label': r'$r$',
                },
                'title': 'Corr(Prediction, Truth)',
            })
            corr = xr.corr(self.da_pred, self.da_truth, dim='time')
            corr.x.plot(ax=ax['corr'], **_kws_corr)

            _kws_R2 = kws.copy()
            _kws_R2.update({
                'levels': np.linspace(0, 1, 11),
                'cbar_kwargs': {
                    'ticks': np.linspace(0, 1, 11),
                    'label': r'R$^2$',
                },
                'extend': 'min',
                'title': r'R$^2$(Prediction, Truth)',
            })
            R2 = xr.apply_ufunc(
                r2_score,
                self.da_pred,
                self.da_truth,
                input_core_dims=[['time'], ['time']],  # Apply along the 'time' dimension
                vectorize=True,
            )
            R2.x.plot(ax=ax['R2'], **_kws_R2)

            _kws_RMSE = kws.copy()
            _kws_RMSE.update({
                'levels': np.linspace(-1, 1, 11),
                'cbar_kwargs': {
                    'ticks': np.linspace(-1, 1, 11),
                    'label': 'RMSE',
                },
                'extend': 'both',
                'title': 'RMSE(Prediction, Truth)',
            })
            MSE = xr.apply_ufunc(
                mean_squared_error,
                self.da_pred,
                self.da_truth,
                input_core_dims=[['time'], ['time']],  # Apply along the 'time' dimension
                vectorize=True,
            )
            RMSE = np.sqrt(MSE)
            RMSE.x.plot(ax=ax['RMSE'], **_kws_RMSE)

        return fig, ax


import xarray as xr
import pybaywatch as pb
import numpy as np

from . import utils
from . import obs

class IdenticalTS:
    def __init__(self, record:obs.ProxyRecord=None):
        self.record = record

    @property
    def clim_vns(self):
        return ['TS', 'tas']

    def forward(self):
        if 'TS' in self.clim:
            output = self.clim['TS'].values
        elif 'tas' in self.clim:
            output = self.clim['tas'].values

        return output

class IdenticalSST:
    def __init__(self, record:obs.ProxyRecord=None):
        self.record = record

    @property
    def clim_vns(self):
        return ['TEMP', 'SST', 'sst', 'tos']

    def forward(self):
        if 'TEMP' in self.clim:
            if 'z_t' in self.clim['TEMP'].dims:
                output = self.clim['TEMP'].isel(z_t=0).values
            else:
                output = self.clim['TEMP'].values

        elif 'SST' in self.clim:
            output = self.clim['SST'].values
        elif 'sst' in self.clim:
            output = self.clim['sst'].values
        elif 'tos' in self.clim:
            output = self.clim['tos'].values
        
        return output

class IdenticalSSS:
    def __init__(self, record:obs.ProxyRecord=None):
        self.record = record

    @property
    def clim_vns(self):
        return ['SALT', 'SSS', 'sss', 'sos']

    def forward(self):
        if 'SALT' in self.clim:
            if 'z_t' in self.clim['SALT'].dims:
                output = self.clim['SALT'].isel(z_t=0).values
            else:
                output = self.clim['SALT'].values
        elif 'SSS' in self.clim:
            output = self.clim['SSS'].values
        elif 'sss' in self.clim:
            output = self.clim['sss'].values
        elif 'sos' in self.clim:
            output = self.clim['sos'].values

        return output

class IdenticalSSTSSS:
    def __init__(self, record:obs.ProxyRecord=None):
        self.record = record

    @property
    def clim_vns(self):
        return ['TEMP', 'SALT']

    def forward(self):
        output = self.clim['TEMP'].isel(z_t=0).values+self.clim['SALT'].isel(z_t=0).values
        return output

class TEX86:
    def __init__(self, record:obs.ProxyRecord=None):
        self.record = record

    @property
    def clim_vns(self):
        return ['TEMP', 'tos', 'sst']

    def forward(self, seed=2333, mode='analog', type='SST', tolerance=1):
        if 'TEMP' in self.clim:
            sst = self.clim['TEMP'].isel(z_t=0).values
        elif 'tos' in self.clim:
            sst = self.clim['tos'].values
        elif 'sst' in self.clim:
            sst = self.clim['sst'].values

        lat = self.record.data.lat
        lon = self.record.data.lon
        lon180 = utils.lon180(lon)

        # run
        self.params = {
            'lat': lat,
            'lon': lon180,
            'temp': sst,
            'seed': seed,
            'type': type,
            'mode': mode,
            'tolerance': tolerance,
        }
        res = pb.TEX_forward(**self.params)
        if res['status'] == 'FAIL':
            utils.p_warning(f'>>> Forward modeling failed for proxy: {self.meta["pid"]}')
            output = None
        else:
            output = np.median(res['values'], axis=1)

        return output

class UK37:
    def __init__(self, record:obs.ProxyRecord=None):
        self.record = record

    @property
    def clim_vns(self):
        return ['TEMP', 'tos', 'sst']

    def forward(self, order=3, seed=2333):
        if 'TEMP' in self.clim:
            sst = self.clim['TEMP'].isel(z_t=0).values
        elif 'tos' in self.clim:
            sst = self.clim['tos'].values
        elif 'sst' in self.clim:
            sst = self.clim['sst'].values

        # run
        self.params = {
            'sst': sst,
            'order': order,
            'seed': seed,
        }
        res = pb.UK_forward(**self.params)
        output = np.median(res['values'], axis=1)
        return output

class MgCa:
    def __init__(self, record:obs.ProxyRecord=None):
        self.record = record

    @property
    def clim_vns(self):
        return ['TEMP', 'tos', 'sst', 'SALT', 'sos', 'sss']

    def forward(self, age, omega=None, pH=None, clean=None, species=None, sw=2, H=1, seed=2333):
        if 'TEMP' in self.clim and 'SALT' in self.clim:
            sst = self.clim['TEMP'].isel(z_t=0).values
            sss = self.clim['SALT'].isel(z_t=0).values
        elif 'tos' in self.clim and 'sos' in self.clim:
            sst = self.clim['tos'].values
            sss = self.clim['sos'].values
        elif 'sst' in self.clim and 'sss' in self.clim:
            sst = self.clim['sst'].values
            sss = self.clim['sss'].values

        # get omega and pH
        lat = self.record.data.lat
        lon = self.record.data.lon
        depth = self.record.data.depth
        if omega is None and pH is None:
            lon180 = np.mod(lon + 180, 360) - 180
            omega, pH = pb.core.omgph(lat, lon180, depth)

        if clean is None: clean = self.record.data.clean
        if species is None: species = self.record.data.species

        # run
        self.params = {
            'age': age,
            'sst': sst,
            'salinity': sss,
            'pH': pH,
            'omega': omega,
            'species': species,
            'clean': clean,
            'sw': sw,
            'H': H,
            'seed': seed,
        }
        res = pb.MgCa_forward(**self.params)
        output = np.median(res['values'], axis=1)
        return output

class d18Oc:
    def __init__(self, record:obs.ProxyRecord=None):
        self.record = record

    @property
    def clim_vns(self):
        return ['TEMP', 'tos', 'sst', 'SALT', 'sos', 'sss', 'd18Osw']

    def forward(self, pH=None, species=None, pH_type=0, seed=2333):
        if 'TEMP' in self.clim and 'SALT' in self.clim:
            sst = self.clim['TEMP'].isel(z_t=0).values
            sss = self.clim['SALT'].isel(z_t=0).values
        elif 'tos' in self.clim and 'sos' in self.clim:
            sst = self.clim['tos'].values
            sss = self.clim['sos'].values
        elif 'sst' in self.clim and 'sss' in self.clim:
            sst = self.clim['sst'].values
            sss = self.clim['sss'].values

        d18Osw = self.clim['d18Osw'].values
        if pH is None: pH = self.record.data.species
        if species is None: species = self.record.data.species

        # run
        self.params = {
            'sst': sst,
            'd18Osw': d18Osw,
            'sss': sss,
            'pH': pH,
            'pH_type': pH_type,
            'species': species,
            'seed': seed,
        }
        res = pb.d18Oc_forward(**self.params)
        output = np.median(res['values'], axis=1)
        return output


import x4c
import numpy as np

class Restart:
    def __init__(self, ds, ds_h, comp, grid):
        self.ds = ds  # restart file
        self.ds_h = ds_h  # hist file or a timeseries file
        self.ds.attrs['comp'] = comp
        self.ds.attrs['grid'] = grid

        # modify the coords
        if comp == 'ocn':
            ds_tmp = self.ds.rename({'i': 'nlon', 'j': 'nlat', 'k': 'z_t'})
        else:
            raise ValueError('Unknown `comp`!')
        
        ds_tmp = ds_tmp.assign_coords(self.ds_h.coords)
        self.ds = ds_tmp
        del(ds_tmp)

        for vn in self.ds.data_vars:
            self.ds[vn].attrs['comp'] = comp
            self.ds[vn].attrs['grid'] = grid
            if comp == 'ocn':
                # mask the NaN area
                mask = self.ds_h['TEMP'].isnull().squeeze()
                self.ds[vn] = self.ds[vn].where(~mask, other=np.nan)

    def __getitem__(self, vn):
        return self.ds[vn]


import numpy as np
import xarray as xr
import x4c
from scipy.linalg import cholesky, sqrtm
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

from multiprocessing import Pool
from functools import partial
from tqdm import tqdm

from sklearn.metrics import (
    r2_score,
    mean_squared_error,
    mean_absolute_error,
    mean_absolute_percentage_error,
)

from .. import utils

def gaspari_cohn(dist, loc_radius):
    '''
    Vectorized Gaspari-Cohn localization function.
    
    Args:
        dist (ndarray): Distance(s) between model state and observation.
        loc_radius (float): Localization radius (distance beyond which covariance is set to zero).

    Reference:
        Gaspari, G., Cohn, S.E., 1999. Construction of correlation functions in two and three dimensions.
        Quarterly Journal of the Royal Meteorological Society 125, 723-757. https://doi.org/10.1002/qj.49712555417
    '''
    # Normalize the distances
    r = np.abs(dist) / loc_radius
    
    # Initialize the result array with zeros
    f = np.zeros_like(r)
    
    # Eq. (4.10) in Gaaspari & Coh (1999)
    mask1 = r <= 1
    f[mask1] = -r[mask1]**5 / 4 + r[mask1]**4 / 2 + 5/8 * r[mask1]**3 - 5/3 * r[mask1]**2 + 1
    
    mask2 = (r > 1) & (r <= 2)
    f[mask2] = r[mask2]**5 / 12 - r[mask2]**4 / 2 + 5/8 * r[mask2]**3  + 5/3 * r[mask2]**2 - 5 * r[mask2] + 4 - 2/3 / r[mask2]

    f[f<0] = 0 # force f >= 0
    return f

def gaspari_cohn_DASH(dist, loc_radius, scale=0.5):
    """
    Implements a Gaspari-Cohn 5th order polynomial localization function following DASH.
    
    Parameters:
        dist (ndarray): An array of distances.
        loc_radius (float): The cutoff radius, beyond which weights are zero.
        scale (float or str, optional): The length scale for the polynomial.
            Must be on the interval 0 < scale <= 0.5, or 'optimal' to use the optimal
            length scale as described by Lorenc (2003). Default is 0.5.
    
    Returns:
        weights (ndarray): Covariance localization weights with the same shape as distances.
    """
    # Set the scale if 'optimal' is specified
    if isinstance(scale, str) and scale == 'optimal':
        scale = np.sqrt(10 / 3)
    
    # Define length scale and localization radius
    c = scale * loc_radius
    
    # Preallocate weights array with ones
    weights = np.ones_like(dist)
    
    # Calculate mask arrays for the different distance ranges
    outside_radius = dist > loc_radius
    inside_scale = dist <= c
    in_between = ~inside_scale & ~outside_radius

    # Apply Gaspari-Cohn polynomial
    X = dist / c
    weights[outside_radius] = 0
    weights[in_between] = X[in_between]**5 / 12 - 0.5 * X[in_between]**4 + 0.625 * X[in_between]**3 + (5 / 3) * X[in_between]**2 - 5 * X[in_between] + 4 - 2 / (3 * X[in_between])
    weights[inside_scale] = -0.25 * X[inside_scale]**5 + 0.5 * X[inside_scale]**4 + 0.625 * X[inside_scale]**3 - (5 / 3) * X[inside_scale]**2 + 1
    
    # Ensure weights are non-negative due to rounding errors
    weights[weights < 0] = 0
    
    return weights



class EnSRF:
    def __init__(self, X=None, Y=None, y=None, R=None, L=None, Lobs=None):
        self.X = X            # ensemble of the prior state vectors (n x N)
        self.Y = Y            # ensemble of the forward estimates (m x N); Y=H(X)
        self.y = y            # observations (m x 1)
        self.R = R            # obs err matrix (m x m)
        self.L = L            # localization matrix (n x m)
        self.Lobs = Lobs      # localization matrix (m x m)

    def update(self, debug=False):
        ''' Perform an EnSRF update with localization. '''
        N = self.X.shape[1]  # Ensemble size

        # Compute the ensemble mean
        Xm = np.mean(self.X, axis=1, keepdims=True)
        Xp = self.X - Xm

        Ym = np.mean(self.Y, axis=1, keepdims=True)
        Yp = self.Y - Ym

        # Observation error covariance matrix
        Ycov = (Yp @ Yp.T) / (N - 1)

        # Localize the obs err covariance matrix
        if self.Lobs is not None:
            Ycov_loc = Ycov * self.Lobs
        else:
            Ycov_loc = Ycov

        C =  Ycov_loc + self.R

        # Kalman gain matrix
        XYcov = (Xp @ Yp.T) / (N - 1)

        # Localize the Kalman gain
        if self.L is not None:
            XYcov_loc = XYcov * self.L
        else:
            XYcov_loc = XYcov

        K = XYcov_loc @ np.linalg.inv(C)

        # Observation innovation
        d = self.y - Ym

        # Update the ensemble mean
        Xm_updated = Xm + K @ d

        # Update the ensemble perturbations
        self.T = np.eye(N) - (Yp.T @ np.linalg.inv(C)) @ Yp / (N - 1)
        Xp_updated = Xp @ self.T

        # Combine updated mean and perturbations
        self.X_updated = Xm_updated + Xp_updated

        if debug:
            self.Xm = Xm
            self.Xp = Xp
            self.Xm_updated = Xm_updated
            self.Xp_updated = Xp_updated
            self.Ym = Ym
            self.Yp = Yp
            self.C = C
            self.K = K
            self.d = d
            self.Ycov = Ycov
            self.Ycov_loc = Ycov_loc
            self.XYcov = XYcov
            self.XYcov_loc = XYcov_loc

    def plot_T(self, cmap='viridis'):
        ''' Check the T matrix
        '''
        data = self.T.T
        nx, ny = np.shape(data)

        # Define color levels
        vmin, vmax = 0, 1
        num_colors = 11
        levels = np.linspace(vmin, vmax, num_colors)

        # Create a discrete colormap
        cmap = plt.get_cmap(cmap, num_colors)
        norm = mcolors.BoundaryNorm(levels, cmap.N)
        cmap.set_under('k')
        cmap.set_over('red')

        # Create X, Y grid for pcolormesh
        x = np.arange(nx + 1)
        y = np.arange(ny + 1)
        X, Y = np.meshgrid(x, y)

        # Create the plot
        fig, ax = plt.subplots()
        pcm = ax.pcolormesh(X, Y, data.T, cmap=cmap, norm=norm, shading='auto')

        # Customize colorbar
        cbar = plt.colorbar(pcm, ax=ax, extend="both")
        cbar.set_ticks(levels)
        cbar.set_label('T Matrix')

        ax.set_title(f'min={data.min():.2f}, max={data.max():.2f}')
        return fig, ax

class EnSRF_DASH:
    def __init__(self, X=None, Y=None, y=None, R=None, L=None, Lobs=None):
        self.X = X            # ensemble of the prior state vectors (n x N)
        self.Y = Y            # ensemble of the forward estimates (m x N); Y=H(X)
        self.y = y            # observations (m x 1)
        self.R = R            # obs err matrix (m x m)
        self.L = L            # localization matrix (n x m)
        self.Lobs = Lobs      # localization matrix (m x m)

    def update(self, debug=False):
        ''' Perform an EnSRF update with localization. '''
        N = self.X.shape[1]  # Ensemble size

        # Compute the ensemble mean
        Xm = np.mean(self.X, axis=1, keepdims=True)
        Xp = self.X - Xm

        Ym = np.mean(self.Y, axis=1, keepdims=True)
        Yp = self.Y - Ym

        # Observation error covariance matrix
        Ycov = (Yp @ Yp.T) / (N - 1)

        # Localize the obs err covariance matrix
        if self.Lobs is not None:
            Ycov_loc = self.Lobs * Ycov
        else:
            Ycov_loc = Ycov

        C =  Ycov_loc + self.R

        # Kalman gain matrix
        XYcov = (Xp @ Yp.T) / (N - 1)

        # Localize the Kalman gain
        if self.L is not None:
            XYcov_loc = self.L * XYcov
        else:
            XYcov_loc = XYcov

        K = XYcov_loc @ np.linalg.inv(C)

        # Observation innovation
        d = self.y - Ym

        # Update the ensemble mean
        Xm_updated = Xm + K @ d

        # Update the ensemble perturbations
        Ksqrt = sqrtm(C)
        Ksqrt_inv_transpose = np.linalg.inv(Ksqrt).T
        Rcov_sqrt = sqrtm(self.R)
        Ka = K @ Ksqrt_inv_transpose @ np.linalg.inv(Ksqrt + Rcov_sqrt)
        Xp_updated = Xp - Ka @ Yp

        # Combine updated mean and perturbations
        self.X_updated = Xm_updated + Xp_updated

        if debug:
            self.Xm = Xm
            self.Xp = Xp
            self.Ym = Ym
            self.Yp = Yp
            self.C = C
            self.K = K
            self.d = d


class EnOI:
    def __init__(self, X_target=None, X=None, Y=None, y=None, R=None, L=None):
        self.X_target = X_target   # the **monthly** prior state vectors (n x 1)
        self.X = X         # ensemble of the prior state vectors (n x N)
        self.Y = Y         # ensemble of the forward estimates (m x N); Y=H(X)
        self.y = y         # observations (m x 1)
        self.R = R         # obs err matrix (m x m)
        self.L = L         # localization matrix (n x m)

    def update(self, debug=False):
        ''' Perform an EnOI update with localization. '''
        N = self.X.shape[1]  # Ensemble size

        # Compute the ensemble mean
        Xm = np.mean(self.X, axis=1, keepdims=True)
        Xp = self.X - Xm

        Ym = np.mean(self.Y, axis=1, keepdims=True)
        Yp = self.Y - Ym

        # Observation error covariance matrix
        C = (Yp @ Yp.T) / (N - 1) + self.R

        # Kalman gain matrix
        K = (Xp @ Yp.T) / (N - 1) @ np.linalg.inv(C)

        # Localize the Kalman gain
        if self.L is not None:
            K_loc = K * self.L
        else:
            K_loc = K

        # Observation innovation
        d = self.y - Ym

        # the increment
        inc = K_loc @ d

        # update
        self.X_target_updated = self.X_target + inc

        if debug:
            self.Xm = Xm
            self.Xp = Xp
            self.Ym = Ym
            self.Yp = Yp
            self.C = C
            self.K = K
            self.K_loc = K_loc
            self.d = d


class Solver:
    def __init__(self, prior=None, obs=None, prior_target=None):
        self.prior = prior.copy()
        self.obs = obs
        self.prior_target = prior_target

    def prep(self, recon_period=None, recon_season=list(range(1, 13)),
             localize=True, loc_method='gaspari_cohn', loc_radius=2500, dist_vsf=1,
             startover=False, get_clim_kws={}, **fwd_kws):
        ''' Prepare Y=H(X) and the localization matrix for DA

        Args:
            dist_vsf (float, list of float): the vertical scaling factor of the distance

        '''
        if recon_period is None:
            self.time = self.obs.ds.time
        else:
            self.time = self.obs.ds.time.sel(time=slice(recon_period[0], recon_period[-1]))

        if startover or not hasattr(self.prior, 'Y'):
            utils.p_header('>>> Proxy System Modeling: Y = H(X)')
            self.obs.get_clim(self.prior, **get_clim_kws)
            self.obs.get_pseudo(**fwd_kws)
            self.prior.get_Y(self.obs)

        if 'month' in self.prior.ds.dims and recon_season is not None:
            utils.p_header(f'>>> Annualizing prior w/ season: {recon_season}')
            self.prior.ds = self.prior.annualize(months=recon_season)

        if localize and (startover or not hasattr(self.prior, 'dist')):
            loc_func = {
                'gaspari_cohn': gaspari_cohn,
                'gaspari_cohn_DASH': gaspari_cohn_DASH,
            }
            utils.p_header('>>> Computing the localization matrix')
            self.prior.get_dist(self.obs, s=dist_vsf)
            self.L = loc_func[loc_method](self.prior.dist, loc_radius)
            self.obs.get_dist()
            self.Lobs = loc_func[loc_method](self.obs.dist, loc_radius)
        else:
            self.L = None
            self.Lobs = None

    def run_t(self, t, method='EnSRF', debug=False):
        algo = {
            'EnSRF': EnSRF,
            'EnSRF_DASH': EnSRF_DASH,
            'EnOI': EnOI,
        }

        idx = list(self.obs.ds.time).index(t)
        kws = {}
        for m in algo.keys():
            kws[m] = {
                'X': self.prior.X,
                'Y': self.prior.Y,
                'y': self.obs.y[idx][..., np.newaxis],
                'R': self.obs.R,
                'L': self.L,
                'Lobs': self.Lobs,
            }

        if self.prior_target is not None:
            kws['EnOI']['X_target'] = self.prior_target.X

        S = algo[method](**kws[method])
        S.update(debug=debug)

        # utils.p_header('>>> Formatting the posterior')
        if method in ['EnSRF', 'EnSRF_DASH']:
            post = utils.states2ds(S.X_updated, self.prior.ds)
        elif method == 'EnOI':
            post = utils.states2ds(S.X_target_updated, self.prior_target.ds)
        
        return S, post

    def run(self, method='EnSRF', debug=False, nproc=1, chunksize=10):
        utils.p_header('>>> DA update')
        if nproc == 1:
            res = [self.run_t(t, method=method, debug=debug) for t in tqdm(self.time, desc='Updating time slices')]
        else:
            run_t_partial = partial(self.run_t, method=method, debug=debug)
            with Pool(processes=nproc) as p:
                res = list(tqdm(p.imap_unordered(run_t_partial, self.time, chunksize=chunksize), total=len(self.time), desc='Updating time slices'))

        S, post = zip(*res)

        self.S = S
        self.post = xr.concat(post, dim='time').assign_coords({'time': self.time})

    def plot_valid(self, ds_target, ds_prior=None, vn=None, metric='R2', valid_period=None, **plot_kws):
        metric_dict = {
            'R2': r2_score,
            'MSE': mean_squared_error,
            # 'RMSE': mean_squared_error,
            'MAE': mean_absolute_error,
            'MAPE': mean_absolute_percentage_error,
        }
        _plot_kws = {
            'corr': dict(
                cmap='RdBu_r',
                levels=np.linspace(-1, 1, 21),
                cbar_kwargs={'ticks': np.linspace(-1, 1, 6), 'label': '$r$'},
                title='Corr(Posterior, Truth)',
                extend='neither',
            ),
            'R2': dict(
                cmap='RdBu_r',
                levels=np.linspace(0, 1, 21),
                cbar_kwargs={'ticks': np.linspace(0, 1, 6), 'label': '$R^2$'},
                title='$R^2$(Posterior, Truth)',
                extend='min',
            ),
            'MSE': dict(
                cmap='RdBu_r',
                levels=np.linspace(0, 4, 21),
                cbar_kwargs={'ticks': np.linspace(0, 4, 11), 'label': 'MSE'},
                title='MSE(Posterior, Truth)',
                extend='both',
            ),
            'RMSE': dict(
                cmap='RdBu_r',
                levels=np.linspace(0, 2, 21),
                cbar_kwargs={'ticks': np.linspace(0, 2, 11), 'label': 'RMSE'},
                title='RMSE(Posterior, Truth)',
                extend='max',
            ),
            'MAE': dict(
                cmap='RdBu_r',
                levels=np.linspace(0, 2, 21),
                cbar_kwargs={'ticks': np.linspace(0, 2, 11), 'label': 'MAE'},
                title='MAE(Posterior, Truth)',
                extend='both',
            ),
            'MAPE': dict(
                cmap='RdBu_r',
                levels=np.linspace(0, 2, 21),
                cbar_kwargs={'ticks': np.linspace(0, 2, 11), 'label': 'MAPE'},
                title='MAPE(Posterior, Truth)',
                extend='both',
            ),
        }

        if valid_period is None:
            valid_time = self.time
        else:
            valid_time = self.time.sel(time=slice(valid_period[0], valid_period[-1]))
            
        if metric in metric_dict:
            if ds_prior is None:
                valid_map = xr.apply_ufunc(
                    metric_dict[metric],
                    ds_target[vn].sel(time=valid_time),
                    self.post[vn].sel(time=valid_time).mean('ens'),
                    input_core_dims=[['time'], ['time']],  # Apply along the 'time' dimension
                    vectorize=True,
                )
                if metric == 'RMSE': valid_map = np.sqrt(valid_map)
                _plot_kws[metric].update(plot_kws)
                fig, ax = valid_map.x.plot(**_plot_kws[metric])
            else:
                valid_map_prior = xr.apply_ufunc(
                    metric_dict[metric],
                    ds_target[vn].sel(time=valid_time),
                    ds_prior[vn].sel(time=valid_time),
                    input_core_dims=[['time'], ['time']],  # Apply along the 'time' dimension
                    vectorize=True,
                )
                valid_map_post = xr.apply_ufunc(
                    metric_dict[metric],
                    ds_target[vn].sel(time=valid_time),
                    self.post[vn].sel(time=valid_time).mean('ens'),
                    input_core_dims=[['time'], ['time']],  # Apply along the 'time' dimension
                    vectorize=True,
                )
                if metric == 'RMSE':
                    valid_map_prior = np.sqrt(valid_map_prior)
                    valid_map_post = np.sqrt(valid_map_post)

                _plot_kws[metric].update(plot_kws)
                fig, ax = x4c.visual.subplots(
                    nrow=2, ncol=1,
                    ax_loc={'prior': (0, 0), 'post': (1, 0)},
                    projs={'prior': 'Robinson', 'post': 'Robinson'},
                    projs_kws={'prior': {'central_longitude': 180}, 'post': {'central_longitude': 180}},
                    figsize=(12, 6),
                    wspace=0,
                    hspace=0.2,
                )
                _plot_kws_prior = _plot_kws[metric].copy()
                _plot_kws_prior['title'] = _plot_kws_prior['title'].replace('Posterior', 'Prior')
                valid_map_prior.x.plot(ax=ax['prior'], **_plot_kws_prior)
                valid_map_post.x.plot(ax=ax['post'], **_plot_kws[metric])

        elif metric == 'corr':
            if ds_prior is None:
                valid_map = xr.corr(
                    ds_target[vn].sel(time=valid_time),
                    self.post[vn].sel(time=valid_time).mean('ens'),
                    dim='time',
                )
                _plot_kws[metric].update(plot_kws)
                fig, ax = valid_map.x.plot(**_plot_kws[metric])
            else:
                valid_map_prior = xr.corr(
                    ds_target[vn].sel(time=valid_time),
                    ds_prior[vn].sel(time=valid_time),
                    dim='time',
                )
                valid_map_post = xr.corr(
                    ds_target[vn].sel(time=valid_time),
                    self.post[vn].sel(time=valid_time).mean('ens'),
                    dim='time',
                )
                _plot_kws[metric].update(plot_kws)
                fig, ax = x4c.visual.subplots(
                    nrow=2, ncol=1,
                    ax_loc={'prior': (0, 0), 'post': (1, 0)},
                    projs={'prior': 'Robinson', 'post': 'Robinson'},
                    projs_kws={'prior': {'central_longitude': 180}, 'post': {'central_longitude': 180}},
                    figsize=(12, 6),
                    wspace=0,
                    hspace=0.2,
                )
                _plot_kws_prior = _plot_kws[metric].copy()
                _plot_kws_prior.update({'title': 'Corr(Prior, Truth)'})
                valid_map_prior.x.plot(ax=ax['prior'], **_plot_kws_prior)
                valid_map_post.x.plot(ax=ax['post'], **_plot_kws[metric])
        elif metric == 'RMSE':
            da_target = ds_target['tas'].sel(time=valid_time)
            da_post = self.post['tas'].sel(time=valid_time).mean('ens')
            RMSE_post = np.sqrt(((da_target-da_post)**2).mean(dim='time'))
            if ds_prior is not None:
                da_prior = ds_prior['tas'].sel(time=valid_time)
                RMSE_prior = np.sqrt(((da_target-da_prior)**2).mean(dim='time'))

            if ds_prior is None:
                _plot_kws[metric].update(plot_kws)
                fig, ax = RMSE_post.x.plot(**_plot_kws[metric])
            else:
                _plot_kws[metric].update(plot_kws)
                fig, ax = x4c.visual.subplots(
                    nrow=2, ncol=1,
                    ax_loc={'prior': (0, 0), 'post': (1, 0)},
                    projs={'prior': 'Robinson', 'post': 'Robinson'},
                    projs_kws={'prior': {'central_longitude': 180}, 'post': {'central_longitude': 180}},
                    figsize=(12, 6),
                    wspace=0,
                    hspace=0.2,
                )
                _plot_kws_prior = _plot_kws[metric].copy()
                _plot_kws_prior.update({'title': 'RMSE(Prior, Truth)'})
                RMSE_prior.x.plot(ax=ax['prior'], **_plot_kws_prior)
                RMSE_post.x.plot(ax=ax['post'], **_plot_kws[metric])

        elif metric == 'nino3.4':
            da_target = ds_target['tas'].sel(time=valid_time).x.geo_mean('nino3.4')
            da_post = self.post['tas'].sel(time=valid_time).x.geo_mean('nino3.4').mean('ens')
            if ds_prior is not None:
                da_prior = ds_prior['tas'].sel(time=valid_time).x.geo_mean('nino3.4')
                corr_prior = np.corrcoef(da_target.values, da_prior.values)[1, 0]
                # R2_prior = r2_score(da_target.values, da_prior.values)
                RMSE_prior = np.sqrt(mean_squared_error(da_target.values, da_prior.values))

            corr_post = np.corrcoef(da_target.values, da_post.values)[1, 0]
            # R2_post = r2_score(da_target.values, da_post.values)
            RMSE_post = np.sqrt(mean_squared_error(da_target.values, da_post.values))

            fig, ax = da_target.x.plot(color='k', label='Target')
            if ds_prior is not None:
                # da_prior.x.plot(ax=ax, label=f'Prior ($r$={corr_prior:.2f}, $R^2$={R2_prior:.2f}, RMSE={RMSE_prior:.2f})', color='tab:blue')
                da_prior.x.plot(ax=ax, label=f'Prior ($r$={corr_prior:.2f}, RMSE={RMSE_prior:.2f})', color='tab:blue')

            # da_post.x.plot(ax=ax, label=f'Posterior ($r$={corr_post:.2f}, $R^2$={R2_post:.2f}, RMSE={RMSE_post:.2f})', color='tab:orange')
            da_post.x.plot(ax=ax, label=f'Posterior ($r$={corr_post:.2f}, RMSE={RMSE_post:.2f})', color='tab:orange', lw=3)
            ax.legend(bbox_to_anchor=(1, 1))
            ax.set_title(f'NINO3.4')

        return fig, ax


import os
import lightning as L
from .. import utils

class Trainer(L.Trainer):
    def __init__(self, log_dirpath='./logs', name='MLDA', accelerator='gpu', min_epochs=1, max_epochs=1000):
        utils.p_header(f'Name: {name}')

        early_stop = L.pytorch.callbacks.early_stopping.EarlyStopping(
            monitor='valid_loss',  # Metric to monitor
            patience=10,          # Wait 3 epochs without improvement
            mode='min',          # Minimize the monitored metric (e.g., loss)
            verbose=True,        # Log when training stops
            min_delta=0.001      # Optional: Minimum change to qualify as improvement
        )

        ckpt_callback = L.pytorch.callbacks.ModelCheckpoint(
            dirpath=log_dirpath, filename=name,
        )

        # Remove the checkpoint file if it exists
        self.ckpt_fpath = os.path.join(log_dirpath, f'{name}.ckpt')
        utils.p_header(f'Checkpoint path: {os.path.abspath(self.ckpt_fpath)}')

        if os.path.exists(self.ckpt_fpath):
            os.remove(self.ckpt_fpath)
        else:
            pass

        super().__init__(
            accelerator=accelerator, devices=1, strategy='auto',
            min_epochs=min_epochs, max_epochs=max_epochs,
            default_root_dir=log_dirpath, callbacks=[early_stop, ckpt_callback]
        )


import torch
import lightning as L
import matplotlib.pyplot as plt
from matplotlib.colors import BoundaryNorm
import numpy as np

class Dataset(L.LightningDataModule):
    def __init__(self, obs=None, prior_da=None, target_da=None):
        super().__init__()
        self.obs = obs
        self.prior_da = prior_da
        self.target_da = target_da

        # c
        self.prior_ens = torch.tensor(self.prior_da.values, dtype=torch.float32)
        self.obs_lats = torch.tensor(self.obs.df['lat'].values, dtype=torch.float32)
        self.obs_lons = torch.tensor(self.obs.df['lon'].values, dtype=torch.float32)
        self.obs_Rs = torch.tensor(self.obs.df['R'].values, dtype=torch.float32)


    def setup(self, stage=None, train_frac=0.7, val_frac=0.1, test_frac=0.2, batch_size=10, num_workers=4):
        self.train_frac = train_frac
        self.val_frac = val_frac
        self.test_frac = test_frac
        self.batch_size = batch_size
        self.num_workers = num_workers

        self.dataset = []
        for it, t in enumerate(self.target_da.time.values):
            target = self.target_da.sel(time=t).values
            obs_vals = np.array([self.obs.ds[vn].sel(time=t).values for vn in self.obs.ds.data_vars])

            sample = {
                'obs_vals': torch.tensor(obs_vals, dtype=torch.float32),  # x
                'target': torch.tensor(target, dtype=torch.float32),      # y
            }
            self.dataset.append(sample)

        # Compute dataset indices
        ds_idx = list(range(len(self.dataset)))
        test_size = int(len(self.dataset) * self.test_frac)
        train_valid_size = len(self.dataset) - test_size
        
        # Split train+val and test sets
        self.train_valid_idx = ds_idx[:train_valid_size]
        self.test_idx = ds_idx[train_valid_size:]
        
        self.train_valid_set = torch.utils.data.Subset(self.dataset, self.train_valid_idx)
        self.test_set = torch.utils.data.Subset(self.dataset, self.test_idx)
        
        # Further split into train and validation sets
        train_size = int(self.train_frac / (self.train_frac + self.val_frac) * train_valid_size)
        val_size = train_valid_size - train_size
        self.train_set, self.valid_set = torch.utils.data.random_split(self.train_valid_set, [train_size, val_size])

    def train_dataloader(self):
        return torch.utils.data.DataLoader(self.train_set, batch_size=self.batch_size, num_workers=self.num_workers, shuffle=True)

    def val_dataloader(self):
        return torch.utils.data.DataLoader(self.valid_set, batch_size=self.batch_size, num_workers=self.num_workers)

    def test_dataloader(self):
        return torch.utils.data.DataLoader(self.test_set, batch_size=self.batch_size, num_workers=self.num_workers)



# class PseudoClimateDataset(torch.utils.data.Dataset):
#     def __init__(self, E=5, T=100, H=8, W=8, max_proxies=20):
#         self.E, self.T, self.H, self.W = E, T, H, W
#         self.prior = torch.randn(E, T, H, W)
#         self.true_field = self.prior.mean(dim=0) + 0.1 * torch.randn(T, H, W)

#         # Precompute proxy maps and masks with attrition
#         self.proxy_maps = []
#         self.proxy_masks = []

#         for t in range(T):
#             frac_available = (t + 1) / T  # More proxies in recent times
#             num_proxies = int(frac_available * max_proxies)

#             proxy_map = torch.zeros(H, W)
#             proxy_mask = torch.zeros(H, W)

#             for _ in range(num_proxies):
#                 h = torch.randint(0, H, (1,))
#                 w = torch.randint(0, W, (1,))
#                 val = self.true_field[t, h, w] + 0.2 * torch.randn(1)
#                 proxy_map[h, w] = val
#                 proxy_mask[h, w] = 1.0

#             self.proxy_maps.append(proxy_map)
#             self.proxy_masks.append(proxy_mask)

#     def __len__(self):
#         return self.T

#     def __getitem__(self, t):
#         x = (
#             self.prior[:, t],         # [E, H, W]
#             self.proxy_maps[t],       # [H, W]
#             self.proxy_masks[t],      # [H, W]
#         )
#         y = self.true_field[t]        # [H, W]
#         return x, y

# class Dataset(L.LightningDataModule):
#     def __init__(self, E=5, T=100, H=8, W=8, max_proxies=20,
#                  train_frac=0.7, val_frac=0.1, test_frac=0.2,
#                  batch_size=10, num_workers=4):
#         super().__init__()
#         self.batch_size = batch_size
#         self.num_workers = num_workers

#         self.dataset = PseudoClimateDataset(E=E, T=T, H=H, W=W, max_proxies=max_proxies)

#         assert abs(train_frac + val_frac + test_frac - 1.0) < 1e-6, 'Train, val, and test fractions must sum to 1.'
#         self.train_frac = train_frac
#         self.val_frac = val_frac
#         self.test_frac = test_frac

#     def setup(self, stage=None):
#         ds_idx = list(range(len(self.dataset)))
#         test_size = int(len(self.dataset) * self.test_frac)
#         train_valid_size = len(self.dataset) - test_size
        
#         # Split train+val and test sets
#         self.train_valid_idx = ds_idx[:train_valid_size]
#         self.test_idx = ds_idx[train_valid_size:]
#         train_valid_set = torch.utils.data.Subset(self.dataset, self.train_valid_idx)
#         self.test_set = torch.utils.data.Subset(self.dataset, self.test_idx)
        
#         # Further split into train and validation sets
#         train_size = int(self.train_frac / (self.train_frac + self.val_frac) * train_valid_size)
#         val_size = train_valid_size - train_size
#         self.train_set, self.valid_set = torch.utils.data.random_split(train_valid_set, [train_size, val_size])

#     def train_dataloader(self):
#         return torch.utils.data.DataLoader(self.train_set, batch_size=self.batch_size, num_workers=self.num_workers, shuffle=True)

#     def val_dataloader(self):
#         return torch.utils.data.DataLoader(self.valid_set, batch_size=self.batch_size, num_workers=self.num_workers)

#     def test_dataloader(self):
#         return torch.utils.data.DataLoader(self.test_set, batch_size=self.batch_size, num_workers=self.num_workers)

#     def plot(self):
#         try:
#             import x4c
#         except ImportError:
#             raise ImportError('x4c is required for this method. Please install x4c via: pip install x4c-exp.')

#         x4c.set_style('journal_spines', font_scale=1.2)
#         fig, ax = plt.subplots(2, 2, figsize=(10, 9))

#         n_bins = 10
#         # Create discrete colormap from a continuous one
#         cmap = plt.get_cmap('RdBu_r', n_bins)  # 'n_bins' discrete colors from RdBu_r
#         bounds = np.linspace(-1, 1, n_bins + 1)  # edges of bins
#         norm = BoundaryNorm(bounds, n_bins)

#         ds = self.dataset
#         im = ax[0, 0].imshow(ds.true_field[-1], origin='lower', cmap=cmap, norm=norm)
#         ax[0, 0].set_xticks(range(0, ds.W, 1))
#         ax[0, 0].set_yticks(range(0, ds.H, 1))
#         ax[0, 0].set_xticklabels(range(0, ds.W, 1))
#         ax[0, 0].set_yticklabels(range(0, ds.H, 1))
#         ax[0, 0].grid(False)
#         ax[0, 0].set_title('True Field', weight='bold')
#         ax[0, 0].set_xlabel('Lon')
#         ax[0, 0].set_ylabel('Lat')

#         im = ax[0, 1].imshow(ds.prior[-1][0], origin='lower', cmap=cmap, norm=norm)
#         ax[0, 1].set_xticks(range(0, ds.W, 1))
#         ax[0, 1].set_yticks(range(0, ds.H, 1))
#         ax[0, 1].set_xticklabels(range(0, ds.W, 1))
#         ax[0, 1].set_yticklabels(range(0, ds.H, 1))
#         ax[0, 1].grid(False)
#         ax[0, 1].set_title('Prior Member: 0', weight='bold')
#         ax[0, 1].set_xlabel('Lon')
#         ax[0, 1].set_ylabel('Lat')

#         ax[1, 0].imshow(ds.proxy_maps[-1], origin='lower', cmap=cmap, norm=norm)
#         ax[1, 0].set_xticks(range(0, ds.W, 1))
#         ax[1, 0].set_yticks(range(0, ds.H, 1))
#         ax[1, 0].set_xticklabels(range(0, ds.W, 1))
#         ax[1, 0].set_yticklabels(range(0, ds.H, 1))
#         ax[1, 0].grid(False)
#         ax[1, 0].set_title('Proxy Map', weight='bold')
#         ax[1, 0].set_xlabel('Lon')
#         ax[1, 0].set_ylabel('Lat')

#         ax[1, 1].imshow(ds.proxy_masks[-1], origin='lower', cmap=cmap, norm=norm)
#         ax[1, 1].set_xticks(range(0, ds.W, 1))
#         ax[1, 1].set_yticks(range(0, ds.H, 1))
#         ax[1, 1].set_xticklabels(range(0, ds.W, 1))
#         ax[1, 1].set_yticklabels(range(0, ds.H, 1))
#         ax[1, 1].grid(False)
#         ax[1, 1].set_title('Proxy Mask', weight='bold')
#         ax[1, 1].set_xlabel('Lon')
#         ax[1, 1].set_ylabel('Lat')

#         cbar = fig.colorbar(im, ax=ax, shrink=0.7)
#         cbar.set_ticks(bounds)


import torch
import lightning as L
from .. import utils
import matplotlib.pyplot as plt
from matplotlib.colors import BoundaryNorm
import x4c
import numpy as np

class CVAE(L.LightningModule):
    def __init__(self, prior_ens, obs_lats, obs_lons, obs_Rs, target_shape, lats,
                 latent_dim=32, hidden_dim=256, lr=1e-3):
        super().__init__()
        self.save_hyperparameters(ignore=["prior_ens", "obs_lats", "obs_lons", "obs_Rs", "lats"])
        self.weights = torch.cos(torch.deg2rad(torch.tensor(lats))).view(-1, 1)

        # Register buffers (conditions)
        self.register_buffer("prior_ens", prior_ens)
        self.register_buffer("obs_lats", obs_lats)
        self.register_buffer("obs_lons", obs_lons)
        self.register_buffer("obs_Rs", obs_Rs)

        self.n_proxies = obs_lats.shape[0]
        self.target_shape = target_shape  # (lat, lon)
        self.output_dim = target_shape[0] * target_shape[1]

        # Derived condition encoding
        self.condition_dim = self.n_proxies * 4 + prior_ens[0].numel()

        # Encoder
        self.encoder = torch.nn.Sequential(
            torch.nn.Linear(self.n_proxies + self.condition_dim, hidden_dim),
            torch.nn.ReLU(),
            torch.nn.Linear(hidden_dim, hidden_dim),
            torch.nn.ReLU(),
        )
        self.fc_mu = torch.nn.Linear(hidden_dim, latent_dim)
        self.fc_logvar = torch.nn.Linear(hidden_dim, latent_dim)

        # Decoder
        self.decoder = torch.nn.Sequential(
            torch.nn.Linear(latent_dim + self.condition_dim, hidden_dim),
            torch.nn.ReLU(),
            torch.nn.Linear(hidden_dim, hidden_dim),
            torch.nn.ReLU(),
            torch.nn.Linear(hidden_dim, self.output_dim)
        )

        self.lr = lr

    def encode(self, x, cond):
        inp = torch.cat([x, cond], dim=1)
        h = self.encoder(inp)
        return self.fc_mu(h), self.fc_logvar(h)

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        return mu + eps * std

    def decode(self, z, cond):
        inp = torch.cat([z, cond], dim=1)
        return self.decoder(inp)

    def on_fit_start(self):
        self.weights = self.weights.to(self.device)

    def forward(self, obs_vals):
        # Handle NaNs in obs
        obs_mask = ~torch.isnan(obs_vals)
        obs_vals = torch.nan_to_num(obs_vals, nan=0.0)

        # Build condition: [batch, num_proxies * 4]
        bsz = obs_vals.shape[0]
        lat = self.obs_lats.unsqueeze(0).expand(bsz, -1)
        lon = self.obs_lons.unsqueeze(0).expand(bsz, -1)
        Rs = self.obs_Rs.unsqueeze(0).expand(bsz, -1)
        obs_cond = torch.cat([obs_vals, lat, lon, Rs], dim=1)  # [B, P*4]

        # Add prior ensemble mean (same for all samples)
        prior_feat = self.prior_ens.flatten(start_dim=1).mean(0).expand(bsz, -1)  # [B, prior_dim]
        cond = torch.cat([obs_cond, prior_feat], dim=1)  # [B, condition_dim]

        # CVAE steps
        mu, logvar = self.encode(obs_vals, cond)
        z = self.reparameterize(mu, logvar)
        recon = self.decode(z, cond)
        recon = recon.view(-1, *self.target_shape)  # Reshape to [B, lat, lon]
        return recon, mu, logvar

    def training_step(self, batch, batch_idx):
        obs_vals = batch['obs_vals']
        target = batch['target']

        recon, mu, logvar = self.forward(obs_vals)
        target = target.view_as(recon)

        loss_recon = torch.nn.functional.mse_loss(recon, target, reduction='mean')
        loss_recon = (loss_recon * self.weights).mean()
        loss_kl = -0.5 * torch.mean(1 + logvar - mu.pow(2) - logvar.exp())
        loss = loss_recon + loss_kl

        self.log_dict(
            {'train_loss': loss, 'recon_loss': loss_recon, 'kl_loss': loss_kl},
            on_step=True, on_epoch=True, prog_bar=True, sync_dist=True,
        )
        return loss

    def validation_step(self, batch, batch_idx):
        obs_vals = batch['obs_vals']
        target = batch['target']
        recon, mu, logvar = self.forward(obs_vals)
        target = target.view_as(recon)

        loss_recon = torch.nn.functional.mse_loss(recon, target, reduction='mean')
        loss_recon = (loss_recon * self.weights).mean()
        loss_kl = -0.5 * torch.mean(1 + logvar - mu.pow(2) - logvar.exp())
        loss = loss_recon + loss_kl

        self.log_dict(
            {'valid_loss': loss, 'recon_loss': loss_recon, 'kl_loss': loss_kl},
            on_step=True, on_epoch=True, prog_bar=True, sync_dist=True,
        )

        # loss = loss_recon
        # self.log_dict(
        #     {'valid_loss': loss},
        #     on_step=True, on_epoch=True, prog_bar=True, sync_dist=True,
        # )

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=self.lr)